#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Ugly hack to fix a weird incompatibility with VFML and modern compilers.
#include <ctype.h>
extern "C" 
{
    __const unsigned short int **__wrap___ctype_b (void) 
    {
      return __ctype_b_loc();
    }
}

extern "C" {
#include "vfml.h"
}

// Boilerplate
#include <vector>
#include <iostream>
#include <fstream>
using namespace std;
typedef unsigned int uint;


/* 
 * Utility Class: Iterates over all possible values of a set of variables,
 * each with a prespecified range.  Use this if you find it helpful.
 */
class StateIterator
{
public:

    /* Construct an iterator from a vector of variable ranges.
     * The iterator will then iterate over all possible combinations 
     * of their values within their respective ranges. 
     */
    StateIterator(vector<int> ranges)
      : _index(0),
        _ranges(ranges), 
        _currState(ranges.size(), 0), 
        _zero(true)
    { /* NOP */ }


    /* Obtain the current value of the i'th variable. 
     */
    int operator[](int idx) {
        return _currState[idx];
    }


    /* Proceed to the next state combination.  This will wrap-around
     * to zero if there are no legal states left.
     */
    StateIterator& operator++()
    {
        _zero = true;

        // Try each variable in turn...
        for (uint i = 0; i < _ranges.size(); i++) { 

            /* Increment var.  If its value is still legal, we're done!
             * Otherwise, set it to zero and keep going.  This will
             * produce a reasonable ordering, e.g.:
             *   0 0 0 -> 1 0 0 -> 0 1 0 -> 1 1 0 -> 0 0 1 -> etc.
             * for three Boolean variables as we increment over and over
             * again.  If we run out of variables, then we must have
             * wrapped around to the zero state.
             */
            _currState[i]++;
            if (_currState[i] < _ranges[i]) {
                _zero = false;
                break;
            } else {
                _currState[i] = 0;
            }
        }

        // Also increment the index referring to this configuration.
        _index++;
        if (_zero == true) {
            _index = 0;
        }
        return *this;
    }

    // Debugging helper -- print this out to see if something's wrong.
    void print(ostream& out)
    {
        out << "si:";
        for (uint i = 0; i < _currState.size(); i++) {
            out << " " << _currState[i] << "/" << _ranges[i];
        }
        out << endl;
    }

    // Determine if we've wrapped around to zero
    bool isZero() { return _zero; }

    // Provide the index of the current state, allowing us to maintain
    // a correspondence between variable states and positions in a vector.
    long index()  { return _index; }

    // Member variables 
    // (Left public for convenience -- this CSE515, not software engineering!)
    long _index;
    vector<int> _ranges;
    vector<int> _currState;
    bool _zero;
};


int main(int argc, char* argv[])
{
    // Load Bayes net from first command-line argument
    BeliefNet bn = BNReadBIF(argv[1]);
    ExampleSpecPtr spec = BNGetExampleSpec(bn);
    int numVars = ExampleSpecGetNumAttributes(spec);

    /*
     * TODO: Initialize and run belief propagation here...
     *
     * For more helpful functions, including how to access the CPTs, see: 
     *    http://www.cs.washington.edu/dm/vfml/
     * Look at the Module on Bayesian Networks, and BeliefNet.h
     *
     */

    // Output results
    ofstream results("results.txt");
    for (int i = 0; i < numVars; i++ ) {
        BeliefNetNode bnn = BNGetNodeByID(bn, i);
        results << BNNodeGetName(bnn);
        for (int vi = 0; vi < BNNodeGetNumValues(bnn); vi++) {
            results << " " << 1.0; // <-- !!! TODO !!! Real prob goes here!
        }
        results << endl;
    }
    results.close();

    return 0;
}
